import ast
import pathlib
import shlex
import sys
import typing as t
import warnings

from . import toxini


def generate_tox_func(env: toxini.Env) -> t.List[str]:
    """Given a tox env dictionary, returns bash code to run."""
    envdir = pathlib.Path(env["envdir"])
    if not envdir.exists():
        raise RuntimeError(
            f"Can't generate env: path {envdir} does not exist."
            "Make sure tox has been run once to set up all virtualenvs and try "
            "again."
        )

    bindir = envdir / "bin"
    whitelist_externals: t.List[str] = ast.literal_eval(env["whitelist_externals"])
    commands: t.List[t.List[str]] = ast.literal_eval(env["commands"])
    changedir: t.Optional[pathlib.Path] = None if not env[
        "changedir"
    ] else pathlib.Path(env["changedir"])

    lines: t.List[str] = []

    if changedir:
        lines.append(f"pushd {changedir}")

    for command in commands:
        if "setenv" in env and env["setenv"].startswith("SetenvDict: "):
            sed = ast.literal_eval(env["setenv"][12:])
            for k, v in sed.items():
                try:
                    v = v.format(**env)
                    lines.append(shlex.quote(k) + "='" + shlex.quote(v) + "' \\")
                except KeyError:
                    warnings.warn("TODO: handle complex environment variables", Warning)

        bin_command = bindir / command[0]
        if not bin_command.exists():
            if command[0] not in whitelist_externals:
                raise RuntimeError(
                    f"Error generating command: {command}\n"
                    f"{command[0]} not in virtualenv or whitelist_externals."
                )
            command_0 = shlex.quote(command[0])
        else:
            command_0 = shlex.quote(str(bin_command))

        command_line = [command_0] + [shlex.quote(s) for s in command[1:]]
        lines.append(" ".join(command_line))
        lines.append("")

    if changedir:
        lines.append("popd")

    return lines


def create_single_tox_script(ini: toxini.Ini, envs: t.List[str]) -> t.List[str]:
    funcs: t.List[t.Tuple[str, t.List[str]]] = []

    for env_name in envs:
        env = ini.get_env_info(env_name)
        funcs.append((env_name, generate_tox_func(env)))

    return create_bash_script(funcs)


def create_multi_tox_script(envs: t.List[t.Tuple[str, toxini.Env]]) -> t.List[str]:
    funcs: t.List[t.Tuple[str, t.List[str]]] = []

    for env_name, env in envs:
        funcs.append((env_name, generate_tox_func(env)))

    return create_bash_script(funcs)


def create_bash_script(funcs: t.List[t.Tuple[str, t.List[str]]]) -> t.List[str]:
    tail = "tail"
    if sys.platform == "darwin":
        # OSX has a version of tail that's not too great, so switch to
        # `gtail` instead, available from GNU coreutils.
        tail = "gtail"

    lines: t.List[str] = []
    lines += ["set -euo pipefail", ""]

    for index, func in enumerate(funcs):
        func_name, func_body = func
        lines.append(f"function run_env_{index}(){{")
        lines.append("    echo '" + ("-" * 52) + "'")
        lines.append(f"    echo '| {func_name:<48} |'")
        lines.append("    echo '" + ("-" * 52) + "'")
        for l in func_body:
            lines.append("    " + l)
        lines.append("}")
        lines.append("")

    lines += [
        "readonly tmpdir=`mktemp -d`",
        "",
        "finished=0",
        "",
        "function clean_up(){",
        "    if [ $finished -eq 0 ]; then",
        "        for pid in ${pids[*]}; do",
        '            kill "${pid}" &> /dev/null',
        "        done",
        "    fi",
        '    rm -r "${tmpdir}" &> /dev/null || true',
        "}",
        "",
        "trap 'clean_up' HUP INT QUIT TERM EXIT",
        "",
    ]

    for index, _ in enumerate(funcs):
        lines.append(f'run_env_{index} &> "${{tmpdir}}"/{index} &')
        lines.append(f"pids[{index}]=$!")
        lines.append("")

    lines.append("")

    lines += [
        "status=0",
        "index=0",
        "set +e",
        "for pid in ${pids[*]}; do",
        "    if [ $status -eq 0 ]; then",
        "        " + tail + ' -f -n +1 --pid=$pid "${tmpdir}/${index}" &',
        "        tail_pid=$!",
        "        wait $pid",
        "        status=$?",
        "        wait $tail_pid",
        "    else",
        '        kill "${pid}" &> /dev/null',
        "    fi",
        "    index=$(($index + 1))",
        "done",
        "",
        "if [ $status -eq 0 ]; then",
        "    echo '                                    O K   : )'",
        "else",
        "    echo '                          F A I L E D !   :('",
        "fi",
        "",
        "finished=1",
        "exit $status",
    ]
    return lines



import enum
import sys
import typing as t

from . import bash
from . import toxini


USAGE = """
Usage: qtox [-c tox_directory] --envs env1 [env2 env3 ...]

Arguments:
    -c tox_directory    Directory containing tox.ini.
    -e env1 [env2 ...]
                      One or more environments defined inside of a tox.ini
                      file, listed in the order you want qtox to run them (put
                      fastest running tasks, like `pep8` / `flake8`, first for
                      best results).

Examples:

Creates a bash script to run three environments in one tox file:

    qtox -e pep8 py27 p36

Create a bash script to run six environments across two tox files:

    qtox -c acme-lib -e pep8 mypy pytest -c acme-rest-api -e pep8 mypy pytest

"""


class ParseMode(enum.Enum):
    C = "c"
    ENV = "e"
    NONE = None


def parse_cmds(args: t.List[str]) -> t.Optional[t.List[t.Tuple[t.Optional[str], str]]]:
    """Parses arguments, returns list of tox directory, environment pairs."""

    results: t.List[t.Tuple[t.Optional[str], str]] = []

    tox_dir: t.Optional[str] = None

    current_mode = ParseMode.NONE

    for index, arg in enumerate(args):
        if arg.startswith("-"):
            if arg == "-c":
                current_mode = ParseMode.C
                continue
            elif arg == "-e":
                current_mode = ParseMode.ENV
                continue
            else:
                print(f"Agument {index} was not expected: {arg}")
                print(USAGE)
                return None
        else:
            if current_mode == ParseMode.C:
                tox_dir = arg
            elif current_mode == ParseMode.ENV:
                results.append((tox_dir, arg))
            else:
                print("Error: expected `-c` or `-e`.")
                print(USAGE)
                return None

    if not results:
        print(USAGE)
        return None

    return results


def main() -> None:
    args = parse_cmds(sys.argv[1:])
    if not args:
        sys.exit(1)

    tox_inis: t.Dict[t.Optional[str], toxini.Ini] = {}

    def get_ini(c: t.Optional[str]) -> toxini.Ini:
        if c not in tox_inis:
            tox_inis[c] = toxini.get_ini(c)

        return tox_inis[c]

    envs: t.List[t.Tuple[str, toxini.Env]] = []
    for arg in args:
        tox_ini = get_ini(arg[0])
        env = tox_ini.get_env_info(arg[1])
        if arg[0]:
            name = f"{arg[0]} -> {arg[1]}"
        else:
            name = arg[1]
        envs.append((name, env))

    lines = bash.create_multi_tox_script(envs)

    print("\n".join(lines))

    # print(ini.toxinidir)
    # for env_name in args.envs:
    #     env = ini.get_env_info(env_name)
    #     print(f"[{env_name}]")

    #     # print(json.dumps(env, indent=4))
    #     # commands = json.loads(env["commands"])

    #     lines = bash.generate_tox_func(env)
    #     for l in lines:
    #         print(l)

    #     # print(env.envdir)
    #     # print(env.changedir)



import pathlib
import setuptools


simple_description = "Lets you re-run Tox commands faster and in parallel."
try:
    with open(pathlib.Path(__file__).parent / "README.rst") as file:
        long_description = file.read()
except BaseException:
    long_description = simple_description


setuptools.setup(
    name="qtox",
    version="0.0.8",
    author="Tim Simpson",
    description="Lets you re-run Tox commands faster and in parallel.",
    long_description=long_description,
    packages=setuptools.find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=["typing"],
    license="MIT",
    entry_points={"console_scripts": ["qtox = qtox.main:main"]},
)



import configparser
import subprocess
import typing as t


Env = t.NewType("Env", dict)


def _run_tox_showconfig(tox_dir: t.Optional[str]) -> str:
    cmd = ["tox"]
    if tox_dir:
        cmd.extend(["-c", tox_dir])

    cmd.extend(["--showconfig"])
    output = subprocess.check_output(cmd)
    return output.decode("utf-8")


class Ini:
    def __init__(self, content: str) -> None:
        self._config = configparser.RawConfigParser()
        self._config.read_string(content)

    def get_env_info(self, name: str) -> Env:
        result = {k: v for k, v in self._config.items("_top_")}
        for section in [name, f"testenv:{name}"]:
            try:
                items = self._config.items(section)
            except configparser.NoSectionError:
                continue

            for k, v in items:
                result[k] = v

            return Env(result)

        raise ValueError(f"Could not find tox env {name}")

    @property
    def toxinidir(self) -> str:
        return self._config.get(section="_top_", option="toxinidir")


def get_ini(tox_dir: t.Optional[str]) -> Ini:
    content = _run_tox_showconfig(tox_dir)
    # Add section headers so the INI parser will work
    fake_content = f"[_top_]\n{content}"
    return Ini(fake_content)
